#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#define F_CPU 16000000UL
#include <util/delay.h>
#include <stdio.h>
#include "sht21.h"

#define TRX_FRAME_BUFFER(index) (*(volatile uint8_t *)(0x180 + (index)))

// Modify these addresses accordingly before uploading the code on each node
uint8_t nod1_address __attribute__((section(".data"))) = 3;
uint8_t node_address __attribute__((section(".data"))) = 3;

short light, voltage;
float temp, humid;

// Interrupt triggered when the Symbol Counter reaches to 0.
// The code simply blinks a Led, for debugging purposes
ISR(SCNT_CMP1_vect)
{
  PORTB ^= (1<<PB7);
}

uint32_t symbol_threshold = 0x00000000;
void initializeSymbolCounter()
{
  // enable asynchronous mode, with external oscillator (32.768kHz in our case)
  ASSR |= _BV(AS2);

  SCOCR1HH = (symbol_threshold >> 24);
  SCOCR1HL = (symbol_threshold & 0x00ff0000) >> 16;
  SCOCR1LH = (symbol_threshold & 0x0000ff00) >>  8;
  SCOCR1LL = (symbol_threshold & 0x000000ff);
  SCCR0 = _BV(SCEN);
  SCCNTHH = 0x00;
  SCCNTHL = 0x00;
  SCCNTLH = 0x00;
  SCCNTLL = 0x00;

  while (SCSR & _BV(SCBSY));
  // enable compare match 1 IRQ
  SCIRQM = _BV(IRQMCP1);
}

void delaySymbolCounter(uint8_t seconds)
{
  symbol_threshold += (((uint32_t)seconds) * 62500);

  SCOCR1HH = (symbol_threshold >> 24);
  SCOCR1HL = (symbol_threshold & 0x00ff0000) >> 16;
  SCOCR1LH = (symbol_threshold & 0x0000ff00) >>  8;
  SCOCR1LL = (symbol_threshold & 0x000000ff);

  while (SCSR & _BV(SCBSY));
}

void setState(uint8_t state)
{
  TRX_STATE = CMD_FORCE_TRX_OFF;
  TRX_STATE = state;
  while (state != TRX_STATUS_struct.trx_status);
}

// send a short frame with: current node id and the values read from the sensors.
void sendFrame()
{
  setState(CMD_PLL_ON);

  TRX_FRAME_BUFFER(0) = 13;	//length - minimum length is 3
  TRX_FRAME_BUFFER(1) = (node_address & 0xFF00) >> 8;	//idH
  TRX_FRAME_BUFFER(2) = node_address & 0x00FF; //idL
  TRX_FRAME_BUFFER(3) = ((int)temp & 0xFF00) >> 8;	//tempH
  TRX_FRAME_BUFFER(4) = (int)temp & 0x00FF; //tempL
  TRX_FRAME_BUFFER(5) = ((int)humid & 0xFF00) >> 8; //humidH
  TRX_FRAME_BUFFER(6) = (int)humid & 0x00FF; //humidL
  TRX_FRAME_BUFFER(7) = (light & 0xFF00) >> 8; //lightH
  TRX_FRAME_BUFFER(8) = light & 0x00FF; //lightL
  TRX_FRAME_BUFFER(9) = (voltage & 0xFF00) >> 8; //voltH
  TRX_FRAME_BUFFER(10) = voltage & 0x00FF; //voltL

  // start transmission
  TRX_STATE = CMD_TX_START;
}

volatile short receivedId;
volatile short receivedT;
volatile short receivedH;
volatile short receivedL;
volatile short receivedV;

// Interrupt triggered after detecting the and of a transmission
// Here we parse the received data
ISR(TRX24_RX_END_vect)
{
  receivedId = TRX_FRAME_BUFFER(0) << 8;
  receivedId |= TRX_FRAME_BUFFER(1);
  if (receivedId == 1 || receivedId == 2 || receivedId == 3)
  {
    PORTB ^= (1<<PB5);

    receivedT = TRX_FRAME_BUFFER(2) << 8;
    receivedT |= TRX_FRAME_BUFFER(3);

    receivedH = TRX_FRAME_BUFFER(4) << 8;
    receivedH |= TRX_FRAME_BUFFER(5);

    receivedL = TRX_FRAME_BUFFER(6) << 8;
    receivedL |= TRX_FRAME_BUFFER(7);

    receivedV = TRX_FRAME_BUFFER(8) << 8;
    receivedV |= TRX_FRAME_BUFFER(9);
  }
  setState(CMD_RX_ON);
}

void rfInit(void)
{
  setState(CMD_TRX_OFF);

  IRQ_STATUS = 0xff;
  IRQ_MASK_struct.rx_end_en = 1;
  sei();

  TRX_CTRL_2_struct.rx_safe_mode = 1;
}

int main(void)
{	
  twiInit();

  sei();

  DDRE = 0xff;
  PORTE = 0xff;
  PRR0 &= ~_BV(PRADC);
  DDRB = 0xff;
  PORTB = 0x00;

  initializeSymbolCounter();
  rfInit();

  // Init UART for node 0 (the gateway)
  if (node_address == 0)
    Serial.begin(9600);

  while(1)
  {
    // Node 0 is the gateway
    if (node_address != 0)
    {
      // Sleep for 1 second
      delaySymbolCounter(1);
      DDRE |= _BV(PE7);
      PORTE &= ~_BV(PE7);
      PORTB |= 0x01;

      _delay_ms(10);

      // Read data from sht21 sensor (humid and temp)
      uint16_t value16 = SHTReadValue(0xE3);
      TWCR = 0;
      temp = ((float)value16) / 374.23 - 46.85;
      value16 = SHTReadValue(0xE5);
      humid = ((float) value16) / 524.288 - 6;	

      // Read voltage value from PF0. For Atmega128RFA1 PF0 is an input for ADC0
      //
      // Enable ADC and set ADC prescalar to 128 - 125KHz sample rate @ 16MHz
      ADCSRA = _BV(ADEN) | _BV(ADPS0) | _BV(ADPS2) | _BV(ADPS1); 
      // Left adjust ADC result to allow easy 8 bit reading and Set ADC reference to AVCC
      ADMUX = _BV(ADLAR) | _BV(REFS0);
      // Poll to see if AVDD has been powered-up.
      loop_until_bit_is_set(ADCSRB, AVDDOK);
      // Start the conversion
      ADCSRA |= _BV(ADSC);
      // Poll to see if A/D conversion is completed
      loop_until_bit_is_set(ADCSRA, ADIF);
      // Valoarea tensiunii citite este divizata cu 2, asa ca va trebui sa o adjustati.
      voltage = ADCH * 2;

      // Read light value from PF2. For Atmega128RFA1 PF2 is an input for ADC2
      //
      // Left adjust ADC result to allow easy 8 bit reading; Set ADC reference to AVCC; Select the input from ADC2
      ADMUX = _BV(ADLAR) | _BV(REFS0) | _BV(MUX1);
      // Poll to see if AVDD has been powered-up.
      loop_until_bit_is_set(ADCSRB, AVDDOK);
      // Start the conversion
      ADCSRA |= _BV(ADSC);
      // Poll to see if A/D conversion is completed
      loop_until_bit_is_set(ADCSRA, ADIF);
      light = ADCH;

      // Send data on the serial port. Useful for debugging non-gateway nodes
      // NOTE: I observed that you can't use a single Serial.begin, at the beggining of main function,
      // and reuse that Serial after waking up from sleep. Junk is printed on the serial in that case.
      // So I use a Serial.begin each time I want to print something and a Serial.end after I finished.
      Serial.begin(9600);
      Serial.print("humid, temp, light, voltage: ");
      Serial.print((int)humid);
      Serial.print(", ");
      Serial.print((int)temp);
      Serial.print(", ");
      Serial.print((int)light);
      Serial.print(", ");
      Serial.println((int)voltage);
      Serial.flush();
      Serial.end();

      // Send the data through the transceiver to a listening gateway
      sendFrame();

      // Sleep and wait to be waken when the symbol counter interrupt is triggered
      DDRE &= ~_BV(PE7);
      PORTE &= ~_BV(PE7);
      PORTB &= 0xFE;	
      set_sleep_mode(SLEEP_MODE_PWR_SAVE);
      sleep_enable();
      sleep_cpu();
      sleep_disable();
    }
    else
    {
      setState(CMD_RX_ON);
      // Do nothing while you have not received a new data package
      while (receivedId != 1 && receivedId != 2 && receivedId != 3) asm("nop");
      //send data on the serial port
      Serial.print((int)receivedId);
      Serial.print(":");
      Serial.print((int)receivedT);
      Serial.print(":");
      Serial.print((int)receivedL);
      Serial.print(":");
      Serial.println((int)receivedH);

      // Blink a Led, for debugging purposes
      PORTB ^= (1<<PB5);
      // Reset receivedId
      receivedId = 0;
    }
  }
}

